import datetime as dt
from typing import Union, List

import pandas as pd

from mindsdb_sql.parser.dialects.mindsdb import CreateJob, DropJob
from mindsdb_sql.parser.ast import Identifier, Star, Select

from mindsdb_sdk.utils.sql import dict_to_binary_op
from mindsdb_sdk.utils.objects_collection import CollectionBase


class Job:
    def __init__(self, project, data):
        self.project = project
        self.data = data
        self._update(data)

    def _update(self, data):
        self.name = data['name']
        self.query_str = data['query']
        self.start_at = data['start_at']
        self.end_at = data['end_at']
        self.next_run_at = data['next_run_at']
        self.schedule_str = data['schedule_str']

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}, query='{self.query_str}')"

    def refresh(self):
        """
        Retrieve job data from mindsdb server
        """
        job = self.project.get_job(self.name)
        self._update(job.data)

    def get_history(self) -> pd.DataFrame:
        """
        Get history of job execution

        :return: dataframe with job executions
        """
        ast_query = Select(
            targets=[Star()],
            from_table=Identifier('jobs_history'),
            where=dict_to_binary_op({
                'name': self.name
            })
        )
        return self.project.api.sql_query(ast_query.to_string(), database=self.project.name)


class Jobs(CollectionBase):
    def __init__(self, project, api):
        self.project = project
        self.api = api

    def list(self, name: str = None) -> List[Job]:
        """
        Show list of jobs in project

        :return: list of Job objects
        """

        ast_query = Select(targets=[Star()], from_table=Identifier('jobs'))

        if name is not None:
            ast_query.where = dict_to_binary_op({'name': name})

        df = self.api.sql_query(ast_query.to_string(), database=self.project.name)

        # columns to lower case
        cols_map = {i: i.lower() for i in df.columns}
        df = df.rename(columns=cols_map)

        return [
            Job(self.project, item)
            for item in df.to_dict('records')
        ]

    def get(self, name: str) -> Job:
        """
        Get job by name from project

        :param name: name of the job
        :return: Job object
        """

        jobs = self.list(name)
        if len(jobs) == 1:
            return jobs[0]
        elif len(jobs) == 0:
            raise AttributeError("Job doesn't exist")
        else:
            raise RuntimeError("Several jobs with the same name")

    def create(self, name: str, query_str: str,
                   start_at: dt.datetime = None, end_at: dt.datetime = None,
                   repeat_str: str = None) -> Union[Job, None]:
        """
        Create new job in project and return it.

        If it is not possible (job executed and not accessible anymore):
           return None

        More info: https://docs.mindsdb.com/sql/create/jobs

        :param name: name of the job
        :param query_str: str, job's query (or list of queries with ';' delimiter) which job have to execute
        :param start_at: datetime, first start of job,
        :param end_at: datetime, when job have to be stopped,
        :param repeat_str: str, optional, how to repeat job (e.g. '1 hour', '2 weeks', '3 min')
        :return: Job object or None
        """

        if start_at is not None:
            start_str = start_at.strftime("%Y-%m-%d %H:%M:%S")
        else:
            start_str = None

        if end_at is not None:
            end_str = end_at.strftime("%Y-%m-%d %H:%M:%S")
        else:
            end_str = None
        ast_query = CreateJob(
            name=Identifier(name),
            query_str=query_str,
            start_str=start_str,
            end_str=end_str,
            repeat_str=repeat_str
        )

        self.api.sql_query(ast_query.to_string(), database=self.project.name)

        # job can be executed and remove it is not repeatable
        jobs = self.list(name)
        if len(jobs) == 1:
            return jobs[0]

    def drop(self, name: str):
        """
        Drop job from project

        :param name: name of the job
        """
        ast_query = DropJob(Identifier(name))

        self.api.sql_query(ast_query.to_string(), database=self.project.name)



from mindsdb_sql.parser.ast import *


def dict_to_binary_op(filters):
    where = None
    for name, value in filters.items():
        where1 = BinaryOperation('=', args=[Identifier(name), Constant(value)])
        if where is None:
            where = where1
        else:
            where = BinaryOperation(
                'and',
                args=[where, where1]
            )
    return where







from functools import wraps
import io

import requests
import pandas as pd

from .. import __about__

def _try_relogin(fnc):
    @wraps(fnc)
    def wrapper(self, *args, **kwargs):
        try:
            return fnc(self, *args, **kwargs)
        except requests.HTTPError as e:
            if e.response.status_code != 401:
                raise e

            # try re-login
            try:
                self.login()
            except requests.HTTPError:
                raise e
            # call once more
            return fnc(self, *args, **kwargs)
    return wrapper


def _raise_for_status(response):
    # show response text in error
    if 400 <= response.status_code < 600:
        raise requests.HTTPError(f'{response.reason}: {response.text}', response=response)


class RestAPI:
    def __init__(self, url=None, login=None, password=None, is_managed=False):

        self.url = url
        self.username = login
        self.password = password
        self.is_managed = is_managed
        self.session = requests.Session()

        self.session.headers['User-Agent'] = f'python-sdk/{__about__.__version__}'

        if login is not None:
            self.login()

    def login(self):
        managed_endpoint = '/api/login'
        cloud_endpoint = '/cloud/login'

        if self.is_managed:
            json = {'password': self.password, 'username': self.username}
            url = self.url + managed_endpoint
        else:
            json = {'password': self.password, 'email': self.username}
            url = self.url + cloud_endpoint
        r = self.session.post(url, json=json)

        # failback when is using managed instance with is_managed=False
        if r.status_code in (405, 404) and self.is_managed is False:
            # try managed instance login

            json = {'password': self.password, 'username': self.username}
            url = self.url + managed_endpoint
            r = self.session.post(url, json=json)

        _raise_for_status(r)

    @_try_relogin
    def sql_query(self, sql, database='mindsdb', lowercase_columns=False):
        url = self.url + '/api/sql/query'
        r = self.session.post(url, json={
            'query': sql,
            'context': {'db': database}
        })
        _raise_for_status(r)

        data = r.json()
        if data['type'] == 'table':
            columns = data['column_names']
            if lowercase_columns:
                columns = [i.lower() for i in columns]
            return pd.DataFrame(data['data'], columns=columns)
        if data['type'] == 'error':
            raise RuntimeError(data['error_message'])
        return None

    @_try_relogin
    def projects(self):
        # TODO not used yet

        r = self.session.get(self.url + '/api/projects')
        _raise_for_status(r)

        return pd.DataFrame(r.json())

    @_try_relogin
    def model_predict(self, project, model, data, params=None, version=None):
        data = data.to_dict('records')

        if version is not None:
            model = f'{model}.{version}'
        if params is None:
            params = {}
        url = self.url + f'/api/projects/{project}/models/{model}/predict'
        r = self.session.post(url, json={
            'data': data,
            'params': params
        })
        _raise_for_status(r)

        return pd.DataFrame(r.json())

    @_try_relogin
    def objects_tree(self, item=''):
        r = self.session.get(self.url + f'/api/tree/{item}')
        _raise_for_status(r)

        return pd.DataFrame(r.json())

    @_try_relogin
    def upload_file(self, name: str, df: pd.DataFrame):

        # convert to file
        fd = io.BytesIO()
        df.to_csv(fd, index=False)
        fd.seek(0)

        url = self.url + f'/api/files/{name}'
        r = self.session.put(
            url,
            data={
                'source': name,
                'name': name,
                'source_type': 'file',
            },
            files={
                'file': fd,
            }
        )
        _raise_for_status(r)

    @_try_relogin
    def upload_byom(self, name: str, code: str, requirements: str):

        url = self.url + f'/api/handlers/byom/{name}'
        r = self.session.put(
            url,
            files={
                'code': code,
                'modules': requirements,
            }
        )
        _raise_for_status(r)

    def status(self) -> dict:

        r = self.session.get(self.url + f'/api/status')
        _raise_for_status(r)

        return r.json()
